﻿using System.Reflection;
using StudioX.Localization.Dictionaries.Xml;

namespace StudioX.Localization.Dictionaries.Json
{
    /// <summary>
    ///     Provides localization dictionaries from JSON files embedded into an <see cref="Assembly" />.
    /// </summary>
    public class JsonEmbeddedFileLocalizationDictionaryProvider : LocalizationDictionaryProviderBase
    {
        private readonly Assembly assembly;
        private readonly string rootNamespace;

        /// <summary>
        ///     Creates a new <see cref="JsonEmbeddedFileLocalizationDictionaryProvider" /> object.
        /// </summary>
        /// <param name="assembly">Assembly that contains embedded json files</param>
        /// <param name="rootNamespace">
        ///     <para>
        ///         Namespace of the embedded json dictionary files
        ///     </para>
        ///     <para>
        ///         Notice : Json folder name is different from Xml folder name.
        ///     </para>
        ///     <para>
        ///         You must name it like this : Json**** and Xml****; Do not name : ****Json and ****Xml
        ///     </para>
        /// </param>
        public JsonEmbeddedFileLocalizationDictionaryProvider(Assembly assembly, string rootNamespace)
        {
            this.assembly = assembly;
            this.rootNamespace = rootNamespace;
        }

        public override void Initialize(string sourceName)
        {
            var resourceNames = assembly.GetManifestResourceNames();
            foreach (var resourceName in resourceNames)
            {
                if (resourceName.StartsWith(rootNamespace))
                {
                    using (var stream = assembly.GetManifestResourceStream(resourceName))
                    {
                        var jsonString = Utf8Helper.ReadStringFromStream(stream);

                        var dictionary = CreateJsonLocalizationDictionary(jsonString);
                        if (Dictionaries.ContainsKey(dictionary.CultureInfo.Name))
                        {
                            throw new StudioXInitializationException(sourceName +
                                                                     " source contains more than one dictionary for the culture: " +
                                                                     dictionary.CultureInfo.Name);
                        }

                        Dictionaries[dictionary.CultureInfo.Name] = dictionary;

                        if (resourceName.EndsWith(sourceName + ".json"))
                        {
                            if (DefaultDictionary != null)
                            {
                                throw new StudioXInitializationException(
                                    "Only one default localization dictionary can be for source: " + sourceName);
                            }

                            DefaultDictionary = dictionary;
                        }
                    }
                }
            }
        }

        protected virtual JsonLocalizationDictionary CreateJsonLocalizationDictionary(string jsonString)
        {
            return JsonLocalizationDictionary.BuildFromJsonString(jsonString);
        }
    }
}